require "defines"

belt_type =
{
    flat = 0,
    down = 1,
    up = 2,
}

to_ground_cost =
{
    ["none"]   =   2,
    ["medium"] =  10,
    ["high"]   = 100,
}

ground_move_cost =
{
    ["none"]   = 2,
    ["medium"] = 2,
    ["high"]   = 5,
}

corner_penalty = 2

function set_pos(tab, pos, val)
    if not tab then print("!tab, ", debug.traceback()) end
    if not pos then print("!pos, ", debug.traceback()) end
    tab[pos.x] = tab[pos.x] or {}
    tab[pos.x][pos.y] = tab[pos.x][pos.y] or {}
    tab[pos.x][pos.y][val] = true
end

--TODO this could be done better
function clear_pos(tab, pos, val)
    if not tab then print("!tab, ", debug.traceback()) end
    if not pos then print("!pos, ", debug.traceback()) end
    tab[pos.x] = tab[pos.x] or {}
    tab[pos.x][pos.y] = tab[pos.x][pos.y] or {}
    tab[pos.x][pos.y][val] = nil
end

function get_pos(tab, pos)
--    print(debug.traceback())
    return tab[pos.x][pos.y]
end

function is_pos(tab, pos, val)
    if not tab then print("!tab, ", debug.traceback()) end
    if not pos then print("!pos, ", debug.traceback()) end
    return tab[pos.x] ~= nil and tab[pos.x][pos.y] ~= nil and tab[pos.x][pos.y][val]
end

function tile_is_passable(surface, pos)
    --this should cover everything we need
    return surface.can_place_entity{name="basic-transport-belt", position=pos}
end

function player_settings(player_index)
    if not global.beltplanner_settings then global.beltplanner_settings = {} end
    local s = global.beltplanner_settings
    if not s[player_index] then
        s[player_index] = {}
        local ps = s[player_index]
        ps.underground_avoidance = "medium"
        ps.underground_length = "short"
        ps.corner_penalty = true
    end
    return s[player_index]
end

function new_minheap()
    function swap_node(a, b)
        local tmp = {element=a.element, value=a.value}
        a.element = b.element
        a.value = b.value
        b.element = tmp.element
        b.value = tmp.value
    end
    result = {
        add = function(self, element, value)
            assert(not is_pos(self.pos, element.pos, element.belt_type))
            set_pos(self.pos, element.pos, element.belt_type)
            self.size = self.size + 1
            local cur_num = self.size
            self.heap[cur_num] = {element=element, value=value, heap_index=cur_num}
            self:percolate(cur_num)
        end,
        contains = function(self, element)
            return is_pos(self.pos, element.pos, element.belt_type)
        end,
        pop = function(self)
            assert(self.size > 0)
            local root = self.heap[1]
            assert(root.heap_index == 1)
            local min_element = root.element
            assert(is_pos(self.pos, min_element.pos, min_element.belt_type))
            clear_pos(self.pos, min_element.pos, min_element.belt_type)
            root.element = self.heap[self.size].element
            root.value = self.heap[self.size].value
            self.heap[self.size] = nil
            self.size = self.size - 1
            self:reheap(1)
            return min_element
        end,
        empty = function(self)
            return self.size == 0
        end,
        get = function(self, element)
            --FIXME ugh...
            for _, v in pairs(self.heap) do
                local a = v.element
                if a.pos.x == element.pos.x and a.pos.y == element.pos.y and a.belt_type == element.belt_type then
                    return v
                end
            end
            return nil
        end,
        reheap = function(self, node)
            local left = 2*node
            local right = left + 1
            local minpos = node
            if left <= self.size and self.heap[left].value < self.heap[minpos].value then
                minpos = left
            end
            if right <= self.size and self.heap[right].value < self.heap[minpos].value then
                minpos = right
            end
            if minpos ~= node then
                swap_node(self.heap[minpos], self.heap[node])
                self:reheap(minpos)
            end
        end,
        percolate = function(self, cur_num)
            local value = self.heap[cur_num].value
            while cur_num > 1 do
                local parent_num = math.floor(cur_num/2)
                assert(self.heap[cur_num].heap_index == cur_num)
                assert(self.heap[parent_num].heap_index == parent_num)
                if self.heap[parent_num].value <= value then
                    break
                else
                    swap_node(self.heap[parent_num], self.heap[cur_num])
                end
                cur_num = parent_num
            end
        end,
        pos = {},
        heap = {},
        size = 0,
    }
    result.__index = result
    return result
end

function get_dir(from, to)
    if from.x == to.x then
        assert(from.y ~= to.y)
        if (from.y > to.y) then return defines.direction.north end
        return defines.direction.south
    end
    assert(from.y == to.y)
    if from.x > to.x then return defines.direction.west end
    return defines.direction.east
end

function find_path(start, goal, player)
    local function h(a, b)
        return math.abs(a.x-b.x) + math.abs(a.y-b.y)
    end

    start.x = math.floor(start.x)
    start.y = math.floor(start.y)
    goal.x = math.floor(goal.x)
    goal.y = math.floor(goal.y)

    local ps = player_settings(player.index)
    local closed = {}
    local open_set = new_minheap()
    open_set:add({pos=start, belt_type=belt_type.flat, g_score=0}, h(start, goal))

    local function neighbors(elem)
        local function valid_neighbor(candidate_pos)
            return not is_pos(closed, candidate_pos, 1)
                and tile_is_passable(player.surface, candidate_pos)
        end
        local new_pos = {}

        if elem.belt_type == belt_type.flat then
            assert(not elem.prev or elem.prev.belt_type ~= belt_type.down)
            for _, dir in pairs({{0, -1}, {0, 1}, {-1, 0}, {1, 0}}) do
                local candidate_pos = { x = elem.pos.x+dir[1], y = elem.pos.y+dir[2] }
                local penalty = 0
                if elem.prev and ps.corner_penalty then
                    local dx = elem.prev.pos.x - elem.pos.x
                    if dx ~= 0 then
                        --we're currently moving horizontally
                        if dir[1] ~= 0 then
                            --keep moving that way with no additional cost
                        else
                            penalty = corner_penalty
                        end
                    else
                        --moving vertically
                        if dir[2] ~= 0 then
                            --great!
                        else
                            penalty = corner_penalty
                        end
                    end
                end
                if valid_neighbor(candidate_pos) then
                    table.insert(new_pos, {pos=candidate_pos,
                                           belt_type=belt_type.flat,
                                           g_score=elem.g_score+1+penalty,
                                           prev=elem})
                    table.insert(new_pos, {pos=candidate_pos,
                                           belt_type=belt_type.down,
                                           dir=dir,
                                           g_score=elem.g_score+to_ground_cost[ps.underground_avoidance]+penalty,
                                           prev=elem})
                end
            end

        elseif elem.belt_type == belt_type.down then
            assert(elem.prev.belt_type ~= belt_type.down)
            for i = 1, 5 do
                local candidate_pos = { x = elem.pos.x+elem.dir[1]*i,
                                        y = elem.pos.y+elem.dir[2]*i }
                if valid_neighbor(candidate_pos) then
                    local penalty
                    if ps.underground_length == "short" then
                        penalty = i * ground_move_cost[ps.underground_avoidance] - 1
                    else
                        penalty = (5 - i) * ground_move_cost[ps.underground_avoidance]
                    end
                    table.insert(new_pos, {pos=candidate_pos,
                                           belt_type=belt_type.up,
                                           dir=elem.dir,
                                           g_score=elem.g_score+penalty,
                                           prev=elem})
                end
            end

        else
            assert(elem.belt_type == belt_type.up)
            assert(elem.prev.belt_type == belt_type.down)
            local candidate_pos = { x = elem.pos.x+elem.dir[1],
                                    y = elem.pos.y+elem.dir[2] }
            if valid_neighbor(candidate_pos) then
                table.insert(new_pos, {pos=candidate_pos,
                                       belt_type=belt_type.flat,
                                       g_score=elem.g_score+1,
                                       prev=elem})
                table.insert(new_pos, {pos=candidate_pos,
                                       belt_type=belt_type.down,
                                       dir=elem.dir,
                                       g_score=elem.g_score+to_ground_cost[ps.underground_avoidance],
                                       prev=elem})
            end
        end

        return function(s, var) s.num=s.num+1 return s.tab[s.num-1] end,
            {num=1, tab=new_pos},
            new_pos[1]
    end

    while not open_set:empty() do
        if open_set.size > 5000 then
            player.print({"openset-too-big"})
            return nil
        end
        local cur = open_set:pop() --this is the one with the lowest f_score
        if cur.pos.x == goal.x and cur.pos.y == goal.y
        and (cur.belt_type == belt_type.flat or cur.belt_type == belt_type.up) then
            local found_path = {}
            table.insert(found_path, {position=goal, direction=get_dir(cur.prev.pos, cur.pos), belt_type=cur.belt_type})
            local next_pos = cur.pos
            cur = cur.prev
            while cur do
                table.insert(found_path, {position=cur.pos, direction=get_dir(cur.pos, next_pos), belt_type=cur.belt_type})
                next_pos = cur.pos
                cur = cur.prev
            end
            return found_path
        end
        set_pos(closed, cur.pos, 1)
        for new_pos in neighbors(cur) do
            if not open_set:contains(new_pos) then
                open_set:add(new_pos, new_pos.g_score + h(new_pos.pos, goal))
            else
                local existing_node = open_set:get(new_pos)
                if existing_node.element.g_score > new_pos.g_score then
                    existing_node.element = new_pos
                    existing_node.value = new_pos.g_score + h(new_pos.pos, goal)
                    open_set:percolate(existing_node.heap_index)
                end
            end
        end
    end
    return nil
end

function create_settings_gui(player)
    if not player.gui.center.beltplanner_config then
        local ps = player_settings(player.index)
        local frame = player.gui.center.add{type="frame",
                                            name="beltplanner_config",
                                            direction="vertical",
                                            caption={"gui-caption"}}
        frame.add{type="label",
                  name="beltplanner_label_underground_avoidance",
                  caption={"label-underground-avoidance"}}
        local flow = frame.add{type="flow", name="beltplanner_avoidance_flow"}
        flow.add{type="checkbox",
                 name="beltplanner_checkbox_avoidance_none",
                 state=(ps.underground_avoidance == "none"),
                 caption={"checkbox-avoidance-none"}}
        flow.add{type="checkbox",
                 name="beltplanner_checkbox_avoidance_medium",
                 state=(ps.underground_avoidance == "medium"),
                 caption={"checkbox-avoidance-medium"}}
        flow.add{type="checkbox",
                 name="beltplanner_checkbox_avoidance_high",
                 state=(ps.underground_avoidance == "high"),
                 caption={"checkbox-avoidance-high"}}
        frame.add{type="label",
                  name="beltplanner_label_length_pre",
                  caption={"label-length-pre"}}
        frame.add{type="checkbox",
                  name="beltplanner_checkbox_short",
                  state=(ps.underground_length == "short"),
                  caption={"checkbox-length-short"}}
        frame.add{type="checkbox",
                  name="beltplanner_checkbox_long",
                  state=(ps.underground_length == "long"),
                  caption={"checkbox-length-long"}}
        frame.add{type="label",
                  name="beltplanner_label_length_post",
                  caption={"label-length-post"}}
        frame.add{type="checkbox",
                  name="beltplanner_checkbox_corner_penalty",
                  state=ps.corner_penalty,
                  caption={"checkbox-corner-penalty"}}
        frame.add{type="button",
                  name="beltplanner_button_close",
                  caption={"button-close"}}
    end
end

script.on_init(function()
    --this also seems to run when the mod is added to an existing save
    global.beltplanner_start = {}
end)

script.on_configuration_changed(function(event)
    if event.mod_changes and event.mod_changes.beltplanner then
        local bp = event.mod_changes.beltplanner
        if not bp.old_version then return end
        --if this is anything else, user must be doing something funny
        assert(bp.old_version == "0.1.0")
        assert(bp.new_version == "0.2.0")
        for _, force in pairs(game.forces) do
            if force.technologies["beltplanner-basic"].researched then
                force.recipes["beltplanner-config"].enabled = true
            end
        end
    end
end)

script.on_event(defines.events.on_built_entity, function(event)
    local e = event.created_entity
    local p = game.get_player(event.player_index)

    --FIXME this is getting awful
    local marker_type = string.match(e.name, "^beltplanner%-(.+)$")
    if marker_type ~= "basic" and marker_type ~= "fast"
    and marker_type ~= "express" and marker_type ~= "end"
    and marker_type ~= "config" then return end

    if marker_type == "config" then
        e.destroy()
        p.cursor_stack.clear() --FIXME doesn't work
        p.insert{name="beltplanner-config"}
        create_settings_gui(p)
        return
    end

    if marker_type == "end" then
        local start = global.beltplanner_start[event.player_index]
        if not start or not start.valid then
            p.print({"missing-start-marker"})
            e.destroy()
            return
        end

        local start_pos = start.position
        local goal_pos = e.position
        local start_name = start.name
        local start_type = string.match(start_name, "^beltplanner%-(.+)$")
        assert(start_type == "basic" or start_type == "fast" or start_type == "express")
        start.destroy()
        global.beltplanner_start[event.player_index] = nil
        e.destroy()
        local path = find_path(start_pos, goal_pos, p)
        if not path then
            p.print({"no-path-found"})
            p.insert{name=start_name, count=1}
            return
        end
        local name
        local type
        for _, v in pairs(path) do
            if v.belt_type == belt_type.flat then
                name = start_type.."-transport-belt"
                type = nil
            elseif v.belt_type == belt_type.down then
                name = start_type.."-transport-belt-to-ground"
                type = "input"
            else
                assert(v.belt_type == belt_type.up)
                name = start_type.."-transport-belt-to-ground"
                type = "output"
            end
            p.surface.create_entity{name="entity-ghost",
                                    inner_name=name,
                                    position=v.position,
                                    direction=v.direction,
                                    force=p.force,
                                    type=type}
        end
        return
    end

    --must have placed a start marker
    local prev_start = global.beltplanner_start[event.player_index]
    if prev_start and prev_start.valid then
        p.print({"multiple-start-markers"})
        p.insert{name=prev_start.name, count=1}
        prev_start.destroy()
    end
    global.beltplanner_start[event.player_index] = e
    local end_count = p.get_item_count("beltplanner-end")
    if end_count > 0 then
        p.remove_item{name="beltplanner-end", count=end_count}
    end
    if p.cursor_stack.valid_for_read and p.cursor_stack.count > 0 then
        assert(p.cursor_stack.name == "beltplanner-"..marker_type)
        local tmp_count = p.cursor_stack.count
        p.cursor_stack.clear()
        p.insert{name="beltplanner-"..marker_type, count=tmp_count}
    end
    p.cursor_stack.set_stack{name="beltplanner-end", count=1}
end)

script.on_event(defines.events.on_player_mined_item, function(event)
    local item = event.item_stack
    assert(item.name ~= "beltplanner-end")

    local marker_type = string.match(item.name, "^beltplanner%-(.+)$")
    if marker_type ~= "basic" and marker_type ~= "fast"
    and marker_type ~= "express" and marker_type ~= "end" then return end

    local p = game.get_player(event.player_index)
    --this isn't necessarily the player who placed the start marker
    --but still, clearing any possible end markers should be the right thing to do
    local count = p.get_item_count("beltplanner-end")
    if count > 0 then
        p.remove_item{name="beltplanner-end", count=count}
    end
end)

script.on_event(defines.events.on_gui_click, function(event)
    local ps = player_settings(event.player_index)
    local ui = event.element.parent

    local name = event.element.name
    if name == "beltplanner_checkbox_avoidance_none" then
        ps.underground_avoidance = "none"
        ui.beltplanner_checkbox_avoidance_none.state = true --prevent this from getting unchecked by clicking again
        ui.beltplanner_checkbox_avoidance_medium.state = false
        ui.beltplanner_checkbox_avoidance_high.state = false
    elseif name == "beltplanner_checkbox_avoidance_medium" then
        ps.underground_avoidance = "medium"
        ui.beltplanner_checkbox_avoidance_none.state = false
        ui.beltplanner_checkbox_avoidance_medium.state = true
        ui.beltplanner_checkbox_avoidance_high.state = false
    elseif name == "beltplanner_checkbox_avoidance_high" then
        ps.underground_avoidance = "high"
        ui.beltplanner_checkbox_avoidance_none.state = false
        ui.beltplanner_checkbox_avoidance_medium.state = false
        ui.beltplanner_checkbox_avoidance_high.state = true
    elseif name == "beltplanner_checkbox_short" then
        ps.underground_length = "short"
        ui.beltplanner_checkbox_short.state = true
        ui.beltplanner_checkbox_long.state = false
    elseif name == "beltplanner_checkbox_long" then
        ps.underground_length = "long"
        ui.beltplanner_checkbox_short.state = false
        ui.beltplanner_checkbox_long.state = true
    elseif name == "beltplanner_checkbox_corner_penalty" then
        ps.corner_penalty = event.element.state
    elseif name == "beltplanner_button_close" then
        ui.destroy()
    end
end)
